#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Premiers pas avec Spark
# 
# Introduction à [Spark](https://spark.apache.org/) et aux [RDD](https://www.cs.cmu.edu/~pavlo/courses/fall2013/static/slides/spark.pdf).

# In[1]:


get_ipython().run_line_magic('matplotlib', 'inline')


# In[2]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Deux ou trois petites choses à ne pas oublier

# ### Local et cluster
# 
# [Spark](http://spark.apache.org/) n'est pas un langage de programmation mais un environnement de calcul distribué. L'installation en locale reproduit ce que [Spark](http://spark.apache.org/) donnerait à grande échelle sur un cluster mais ce n'est pas rigoureusement identique. En particulier cela veut dire que si votre script tourne en local sur un petit jeu de données, il est possible qu'il échoue sur le cluster :
# 
# * Les dépendances du script sont installées en local mais pas sur chaque machine du cluster [Spark](http://spark.apache.org/). Cela peut se faire à l'installation du cluster pour des dépendances conséquentes ou juste avant l'exécution d'un *job* pour des dépendances ponctuelles.
# * Les données sur le cluster sont en plus grand nombre, il est fort probable que l'échantillon aléatoire local ne soit pas représentatif.
# * Les chemins locaux ne fonctionnent pas sur le cluster. Il faudra d'abord uploader les données sur le cluster pour faire tourner le script.
# * Débugger est compliqué : les print ne marchent pas souvent, surtout si c'est en distribué. Le print va s'exécuter sur une machine distance qui est à mille lieues de votre écran.
# 
# Quand ça plante sur une machine distante, il faut s'accrocher. Le pire, c'est quand l'erreur arrive pour une observation toute bizarre après cinq heures de calcul. Si le message d'erreur n'est pas trop incompréhensible, on sen tire. En fait, le plus agaçant, c'est quand le calcul est carrément interrompu par le cluster au bout de cinq heures car il décrète que les probabilités d'aboutir sont quasi nulles. Là, on connaît l'erreur (skewed dataset) et on sait qu'on va souffrir pour construire la contournante.

# ### Spark et RDD
# 
# [Spark](http://spark.apache.org/) ne manipule pas des fichiers mais des [Resilient Distributed Dataset](http://www-bcf.usc.edu/~minlanyu/teach/csci599-fall12/papers/nsdi_spark.pdf) ou *RDD*. En particulier :
# 
# 1. Les *RDD* sont organisés en ligne : ce sont des blocs qui ne seront jamais *cassés* ni *modifié*. Ces lignes ne peuvent pas excéder 2 Go (voir [SPARK-6235](https://issues.apache.org/jira/browse/SPARK-6235)) mais il est conseillé de ne pas aller au-delà de quelques Mo.
# 2. Sauf exception, il est impossible d'accéder à une partie du fichier. Il faut le parcourir en entier (il n'y a pas d'index).
# 3. Les *RDD* fonctionnent comme des *flux* ou *stream*. On peut soit les lire, soit les écrire mais jamais les deux en même temps. Par conséquent, on ne peut pas modifier un *RDD*, il faut toujours en créer un autre.
# 4. Les *RDD* sont distribués. L'ordre des lignes qui le composent n'est pas prévisible.
# 5. Comme l'ordre est imprévisible, on ne stocke **jamais** les noms des colonnes dans les *RDD*.

# ### Les partitions
# 
# Il existe une exception au point 2 : les [partitions](http://dev.sortable.com/spark-repartition/). Une partition est un ensemble de lignes traitées par le même processus. La parallélisation ne peut excéder le nombre de partitions. Par défaut, c'est aléatoire (hash hash). Mais on peut tout-à-fait partionner selon une colonne, deux colonnes. D'ailleurs, c'est là-dessus qu'on joue pour optimiser la distribution. Si on réduit (ou grouper) selon une colonne, c'est d'autant plus rapide si le stream est déjà partitionnée sur cette colonne.

# ### Spark et Python
# 
# Spark est implémenté en Java. L'API [Python](http://spark.apache.org/docs/latest/api/python/pyspark.html#pyspark.RDD) permet de faire beaucoup de choses mais :
# 
# * Elle ne sera jamais aussi complète que l'API [Java](http://spark.apache.org/docs/latest/api/java/index.html?org/apache/spark/api/java/JavaRDD.html).
# * Elle sera plus lente que l'API [Java](http://spark.apache.org/docs/latest/api/java/index.html?org/apache/spark/api/java/JavaRDD.html) ou [Scala](http://spark.apache.org/docs/latest/api/scala/index.html#org.apache.spark.rdd.RDD) (car [Scala](http://spark.apache.org/docs/latest/api/scala/index.html#org.apache.spark.rdd.RDD) est une surcouche fonctionnelle de [Java](http://spark.apache.org/docs/latest/api/java/index.html?org/apache/spark/api/java/JavaRDD.html)).
# 
# 

# ### Librairies sur Spark
# 
# Un des succès de Spark est de proposer des extensions dédiées à certains usages comme [MLlib](http://spark.apache.org/mllib/) qui implémente des algorihmes de machine learning distribués, [GraphX](http://spark.apache.org/graphx/) pour des algorithmes sur des graphes. [MLlib](http://spark.apache.org/mllib/) sera bientôt remplacé par ML qui s'appuie sur les [DataFrame](http://spark.apache.org/docs/latest/sql-programming-guide.html#datasets-and-dataframes).

# ### Erreur : Cannot run program "python"
# 
# Il vous manque probablement ``PYSPARK_PYTHON``.

# In[3]:


import os
for o, v in sorted(os.environ.items()):
    if "SPARK" in o.upper():
        print("{0:25}= {1}".format(o, v.replace(os.environ["USERNAME"], "<username>")))


# ### Erreur : Output directory  file:/... already exists
# 
# Spark n'aime pas écrire des données dans un RDD qui existe déjà. Il faut le supprimer. Tout dépend de l'environnement où on se trouve, sur Hadoop ou en local. Comme c'est en local, nous ferons :

# In[4]:


from pyquickhelper.filehelper import remove_folder
def clean(folder):
    if os.path.exists(folder):
        return remove_folder(folder)
    else:
        return []
clean("fichier.out.txt")


# ### Vérifier que Spark en local fonctionne
# 
# On essaye le *"hello world"* en *Spark* qui consiste à compter les mots dans un fichier. On prend le fichier du notebook.

# In[5]:


text_file = sc.textFile("spark_first_steps.ipynb")
counts = text_file.flatMap(lambda line: line.split(" ")).map(lambda word: (word, 1)).reduceByKey(lambda a, b: a + b)
counts.saveAsTextFile("fichier.out.txt")


# In[6]:


os.listdir("fichier.out.txt/")


# ### Sortie en plusieurs fichiers
# 
# Un *job* *Spark* est distribué. La sortie d'un *job* *Spark* s'effectue sous la forme de plusieurs stream dans un répertoire, un stream par processus. Cela explique la présence de *part-00000*, *part-00001*. Le fichier ``_SUCCESS`` indique le statut du job.

# In[7]:


get_ipython().run_line_magic('load_ext', 'pyensae')
get_ipython().run_line_magic('head', 'fichier.out.txt/part-00000 -n 3')


# Le format dépend du dernier résultat.

# ## Les opérations de bases
# 
# Documentation : [programming-guide.html - transformations](http://spark.apache.org/docs/latest/programming-guide.html#transformations).
# 
# Dans cette section, on considère les données comme un ensemble de lignes de texte. Rien de plus. Donc, pas d'information de type, des conversions quasiment tout le temps. Bref, c'est utile pour comprendre. On y revient quand le reste ne marche pas. En général, on commence par [Spark SQL](http://spark.apache.org/docs/latest/sql-programming-guide.html). Ah oui j'oubliais, on s'en sert beaucoup quand les données ne sont pas structurées et sont décrites par du JSON, genre des logs d'un site internet. Chaque ligne est en fait un gros JSON.
# 
# On utilise un jeu de données de machine learning [Adult](https://archive.ics.uci.edu/ml/datasets/Adult) légèrement pré-traités et que vous pourrez trouver sur GitHub : [td3a_spark](https://github.com/sdpython/ensae_teaching_cs/tree/master/_doc/notebooks/td3a_spark).

# In[8]:


import os
if not os.path.exists("data_adult.txt"):
    from pyquickhelper.filehelper import unzip_files
    unzip_files("data_adult.zip", where_to=".")
assert os.path.exists("data_adult.txt")


# In[9]:


import pandas
df = pandas.read_csv("data_adult.txt", sep="\t", encoding="utf-8")
df.head()


# On enlève le nom des colonnes.

# In[10]:


df.to_csv("adult.txt", sep="\t", encoding="utf-8", index=False, header=None)


# In[11]:


get_ipython().run_line_magic('head', 'adult.txt -n 2')


# ### déclaration d'un RDD
# 
# La déclaration déclare l'existence d'un *RDD* comme on déclare un fichier. Pour l'instant aucune manipulation.

# In[12]:


rdd = sc.textFile("adult.txt")


# ### enregistrement d'un RDD

# In[13]:


import os
if not os.path.exists("out"):
    os.mkdir("out")


# In[14]:


clean("out/copy_adult.txt")
rdd.saveAsTextFile(os.path.abspath("out/copy_adult.txt"))


# In[15]:


get_ipython().run_line_magic('head', 'out/copy_adult.txt/part-00000 -n 2')


# ### lecture locale d'un RDD avec pandas
# 
# On lit chaque morceaux avant de les concaténer.

# In[16]:


import glob
import pandas
def read_rdd(path, **options):
    pat = os.path.join(path, "part*")
    all_files = glob.glob(pat)
    if len(all_files) == 0:
        raise Exception("No file to read in '{0}'".format(path))
    merge = []
    for f in all_files:
        try:
            df = pandas.read_csv(f, header=None, **options)
        except Exception as e:
            raise Exception("Unable to read '{0}'".format(f)) from e
        merge.append(df)
    if len(merge) == 0:
        raise Exception("No file to read in '{0}'".format(path))
    concatenated_df = pandas.concat(merge, ignore_index=True)
    return concatenated_df

data = read_rdd("out/copy_adult.txt", sep="\t", encoding="utf-8")
data.head(n=2)


# ### collect
# 
# Cette opération regroupe les deux précédentes en une seule. Il faut toute de même faire attention de ne pas l'exécuter sur un grand fichier sous peine de faire exploser la mémoire.

# In[17]:


res = rdd.collect()


# In[18]:


res[:2]


# In[19]:


import pandas
df = pandas.DataFrame([_.split("\t") for _ in res])
df.head(2)


# ### map
# 
# Transformer une ligne en une autre ligne. Chaque ligne est traitée indépendemment des autres.

# In[20]:


def extract_column(cols, row):
    spl = row.split("\t")
    return [spl[i].strip() for i in cols]

res = rdd.map(lambda row: extract_column([1,3], row))
res.collect()[:2]


# ### filter
# 
# Garder ou jeter une ligne. Chaque ligne est traitée indépendemment des autres.

# In[21]:


def filter_column(row):
    spl = row.split("\t")
    return spl[-1].strip() != "<=50K"

res = rdd.filter(lambda row: filter_column(row))
res.collect()[:2]


# On combine souvent les deux :

# In[22]:


def filter_column_split(row):
    return row[-1].strip() != "<=50K"

res = rdd.map(lambda row: extract_column([1,3,-1], row)) \
         .filter(lambda row: filter_column_split(row))
res.collect()[:2]


# Il faut faire attention aux transformations successives des lignes.

# ### flatMap
# 
# C'est la principale différence avec SQL. Une ligne peut devenir un nombre variable de lignes.

# In[23]:


def extract_column_and_multiply_row(n, row):
    spl = row.split("\t")
    return [tuple(_.strip() for _ in spl)] * n

res = rdd.flatMap(lambda row: extract_column_and_multiply_row(2, row))
res.collect()[:3]


# ### group / reduce + mapValues
# 
# Petite moyenne ?

# In[24]:


def extract_age_rich(row):
    spl = row.split("\t")
    target = spl[-1].strip()
    age = float(spl[0])
    return (age, target)

def custom_agg(aggset):
    temp = list([_[0] for _ in aggset])
    return len(temp), sum(temp)

ave = rdd.map(extract_age_rich).groupBy(lambda row: row[1]).mapValues(custom_agg)
fin = ave.collect()
fin


# ### sort
# 
# Je n'en parle pas. Trier un gros jeu de données est à proscrire. On peut trier au sein d'un groupe mais **jamais** un stream entier. Ca fait presque dix ans que j'écris des jobs map/reduce, je n'ai jamais écrit un *sort* sur tout un jeu de données. Ca s'appelle flinguer de la CPU pour rien.

# ### join
# 
# Et on remet la moyenne dans le stream initial. Il vaut mieux regarder la documentation de la méthode [join](http://spark.apache.org/docs/latest/api/python/pyspark.html#pyspark.RDD.join) avant de commencer à lire le code qui suit.

# In[25]:


add_key = rdd.map(lambda row: row.split("\t")).map(lambda row: (row[-1].strip(), row))
join = add_key.join(ave)
join.collect()[:2]


# On commence à comprendre pourquoi [Spark SQL](http://spark.apache.org/docs/latest/sql-programming-guide.html), ça risque d'être pas mal.

# ### le choix existentiel du join : le petit join
# 
# On fait souvent une opération qui consiste à garder les lignes pour lesquelles une certaine valeur appartient à un ensemble. On peut faire un join classique ou alors l'ensemble est petit, traiter ce join comme un map. On broadcaste l'ensemble à chaque processus exécutant le map.

# In[26]:


from pyspark.context import SparkContext
ages = sc.broadcast([20, 30, 40])
ages.value


# In[27]:


subset = rdd.filter(lambda row: int(row.split("\t")[0]) in ages.value )
subset.collect()[:2]


# ### les trucs qui servent parfois parce que ... à l'usage ça sert
# 
# Ce que font les méthodes associées aux [RDD](http://spark.apache.org/docs/latest/api/python/pyspark.html#pyspark.RDD), un peu comme les itérateurs, n'est pas toujours intuitif, mais il est à peu près sûr qu'elles vous serviront un jour (peut-être après avoir googlé ou bingé comme des fous).

# In[28]:


simple_rdd = sc.parallelize([2, 3, 4])
simple_rdd.collect()


# In[29]:


simple_rdd.flatMap(lambda x: range(1, x)).collect()


# [histogram](http://spark.apache.org/docs/latest/api/python/pyspark.html#pyspark.RDD.histogram), [groupByKey](http://spark.apache.org/docs/latest/api/python/pyspark.html#pyspark.RDD.groupByKey)

# ### le truc à retenir
# 
# collect, collect... qu'est-ce que je voulais dire déjà... Ah oui... Un job map/reduce c'est :
# 
# 1. La déclaration des flux d'entrées.
# 2. Le traitement à proprement parler.
# 3. La déclaration des flux de sorties.
# 
# A moins d'écrire du java bas niveau, le job est transformé en un plan d'exécution qui n'est jamais exécuté si [collect](http://spark.apache.org/docs/latest/api/python/pyspark.html#pyspark.RDD.collect) ou [save machin chouette](http://spark.apache.org/docs/latest/api/python/pyspark.html#pyspark.RDD.saveAsTextFile) n'est jamais exécuté. Bref, c'est du [lazy](https://en.wikipedia.org/wiki/Lazy_evaluation).

# ## Spark DataFrame
# 
# [Spark SQL](http://spark.apache.org/docs/latest/sql-programming-guide.html)
# 
# Au début, ça commence par... créer un dataframe. Et comme pour pandas, ces objets retienennt les noms et les types.

# In[30]:


import pandas
data = pandas.read_csv("data_adult.txt", sep="\t", encoding="utf-8")
data.head(2)


# In[31]:


if "spark" not in locals():
    from pyspark.sql import SparkSession
    spark = SparkSession.builder.appName("nimportequoi").getOrCreate()  # à ne faire qu'une fois


# In[32]:


# sdf = spark.createDataFrame(data)  # ça marche
sdf = spark.read.csv("data_adult.txt", sep="\t", encoding="utf-8")


# In[33]:


sdf.show()


# ### Conversion à pandas

# In[34]:


df = sdf.toPandas()


# In[35]:


df.head()


# ### Retour aux RDD

# In[36]:


sdf.rdd


# ### Récuperer le schéma

# In[37]:


sdf.schema


# In[38]:


sdf.printSchema()


# ### Utiliser pandas pour spécifer le format
# 
# On utilise pandas sur une partie du stream.

# In[39]:


import pandas
df = pandas.read_csv("data_adult.txt", sep="\t", encoding="utf-8")
df.head(n=2)


# In[40]:


sdf = spark.createDataFrame(df)


# In[41]:


sdf.printSchema()


# In[42]:


fullsdf = spark.createDataFrame(sdf.rdd, sdf.schema)


# In[43]:


fullsdf.printSchema()


# ### Enregistrement au format parquet

# In[44]:


fullsdf.write.parquet("data_adult.schema.parquet")


# ### Relecture du format parquet

# In[45]:


newsdf = spark.read.parquet("data_adult.schema.parquet/")


# In[46]:


newsdf.printSchema()


# ### Dataframe Spark VS Dataframe pandas
# 
# Spark a reproduit la même interface que pandas pour ses dataframes excepté que le résultat n'est pas calculé tant qu'on ne choisit pas de sauvegarder le résultat.

# In[47]:


fifty = fullsdf [fullsdf.age > 50]


# In[48]:


fifty.show()


# In[49]:





# In[50]: